package BaryCentric;

use strict;
use warnings;
use Carp;

# formulas stolen from 
# http://en.wikipedia.org/wiki/Barycentric_coordinates_%28mathematics%29
sub getInterpolationFactors {
	my ($x,$y,$x1,$y1,$x2,$y2,$x3,$y3) = @_;
	my ($z   ,$z1    ,$z2    ,$z3    ) = (0,0,0,0);
	my $A = $x1-$x3;
	my $B = $x2-$x3;
	my $C = $x3-$x;
	my $D = $y1-$y3;
	my $E = $y2-$y3;
	my $F = $y3-$y;
	my ($G,$H,$I) = (0,0,0);
	if ($A == 0 and $B ==0){
		($A,$D) = ($D,$A);
		($E,$B) = ($B,$E);
		($C,$F) = ($F,$C);
	}
	if(($A*($E+$H)-$B*($D+$G)) == 0){
		confess "div by zero!";
	}
	my $l1 = ($B*($F+$I)-$C*($E+$H))/($A*($E+$H)-$B*($D+$G));
	my $l2 = ($A*($F+$I)-$C*($D+$G))/($B*($D+$G)-$A*($E+$H));
	my $l3 = 1-$l1-$l2;
	return ($l1,$l2,$l3);
}

sub smul{
	my ($a,$b) = @_;
	return [$a*$$b[0],$a*$$b[1]];
}
sub padd{
	my ($a,$b) = @_;
	return [$$a[0]+$$b[0],$$a[1]+$$b[1]];
}
sub psub{
	my ($a,$b) = @_;
	return [$$a[0]-$$b[0],$$a[1]-$$b[1]];
}
sub getAddPoints{
	my ($maxmax,$minmin,$maxmin,$minmax) = @_;
	my $p1 = padd($maxmax,psub($maxmax,$minmin));
	my $p2 = padd($minmin,psub($minmin,$maxmax));
	my $p3 = padd($maxmin,psub($maxmin,$minmax));
	my $p4 = padd($minmax,psub($minmax,$maxmin));

	my $p5 = padd($p1,smul(0.5,psub($p3,$p1)));
	my $p6 = padd($p1,smul(0.5,psub($p4,$p1)));

	my $p7 = padd($p2,smul(0.5,psub($p3,$p2)));
	my $p8 = padd($p2,smul(0.5,psub($p4,$p2)));

	return ($p1,$p2,$p3,$p4,$p5,$p6,$p7,$p8);
}

1;
