package RuleBase;
use strict;
use warnings;
use Carp;
use Scalar::Util qw/reftype/;

our $AUTOLOAD;

sub new {
    my ( $pkg, %proto ) = @_;
    delete $proto{__RULE__};    #This information is already in the class name
    bless \%proto, $pkg;
}

sub comment2{
	my $self = shift;
	return "\t/*" . $self->{'comment2'} . "*/" if(exists $self->{'comment2'});
	return '';
}
sub comment{
	my $self = shift;
	return "\t/*" . $self->{'comment'} . "*/" if(exists $self->{'comment'});
	return '';
}
sub rndf{
	my $self = shift;
	my $name = ref $self;
	carp "converting $name to RNDF is not yet implemented.";
	return "";
}
sub optional{ # <<<2
	my $self = shift;
	return $self->comment =~ /optional/i;
}
sub inserted{ # <<<2
	my $self = shift;
	return $self->comment =~ /inserted/i;
}

sub rule {
	my $self = shift;
	confess "rule nonexistent for " . ref($self) unless(exists $self->{__STRING1__});
	return $self->{__STRING1__};
}

sub id1 {
	my $self = shift;
	$self->id =~ m/^(\d+)/ or confess (ref($self)." has no id1");
	return $1
}
sub id2 {
	my $self = shift;
	$self->id =~ m/^(\d+)\.(\d+)/ or confess (ref($self)." has no id2");
	return $2
}
sub id12 {
	my $self = shift;
	$self->id =~ m/^(\d+)\.(\d+)/ or confess (ref($self)." has no id12");
	return "$1.$2"
}
sub id3 {
	my $self = shift;
	$self->id =~ m/^(\d+)\.(\d+)\.(\d+)/ or confess (ref($self)." has no id3");
	return $3
}
sub id123 {
	my $self = shift;
	$self->id =~ m/^(\d+)\.(\d+)\.(\d+)/ or confess (ref($self)." has no id123");
	return "$1.$2.$3"
}
sub id23 {
	my $self = shift;
	$self->id =~ m/^(\d+)\.(\d+)\.(\d+)/ or confess (ref($self)." has no id23");
	return "$2.$3"
}

sub AUTOLOAD{
	my $self = shift;
	my $type = ref($self) or confess "$self is not an object";
	my $name = $AUTOLOAD;
	$name =~ s/.*://;   # strip fully-qualified portion
	unless(exists $self->{$name}){
		if(exists $self->{"$name(s)"}){
			$name = "$name(s)";
		}elsif(exists $self->{"$name(s?)"}){
			$name = "$name(s?)";
		}else{
			confess "Cannot access $name in class $type!";
		}
	}
	if(ref($self->{$name}) and reftype($self->{$name}) eq "HASH" and exists($self->{$name}->{__VALUE__})){
		if(@_){
			return $self->{$name}->{__VALUE__} = shift;
		}else{
			return $self->{$name}->{__VALUE__};
		}
	}
	if(@_){
		return $self->{$name} = shift;
	}else{
		return $self->{$name};
	}
}

sub DESTROY{
}

1;
