#!/usr/bin/perl -w
# vim:fdm=marker:fmr=<<<,>>>
package RndfEditor; # <<<1
use strict;
# random stuff
use List::Util  qw/min max reduce sum/;
use Digest::MD5 qw/md5_hex/;
use Math::NumberCruncher;
use Math::Vec qw(:terse);
use Config::IniFiles;
use Data::Dumper;
use Clone qw(clone);
use File::stat;
use File::Remove;
use Smart::Comments '###';
use Geo::Distance;
use TriangleWrap;

# my stuff
use RNDF;
use TriangleWrap;
use BaryCentric;
use GPST12N;

# GUI
use Gtk2 '-init';
use Gtk2::CV::ImageWindow;
use GD::Simple;
use GD::Convert qw(gif=gif_netpbm newFromGif=newFromGif_imagemagick wbmp);
use Gtk2::CV::Progress;


# Variables <<<2 
my $cfg = Config::IniFiles->new ( -file => "rndf_ed.ini" );
die join(',',@Config::IniFiles::errors) unless $cfg;

my $win;

my %waypoints;      # points from RNDF file
my %wptransformed;  # points from RNDF file, transformed to map coordinate system
my %supportpoints;  # selected points with associated coordinates defined by user

my %checkpoints;
my %exitpoints;
my %insertedwp;
my %optionalwp;
my %centerCalcWP;
my @stoppoints;
my @perimeters;
my @zonelanes;
my @lines;

# these are the numbers determining what we're looking at in the main window
my $rndffile;
my $largefile;
my $large_width;
my $large_height;
my $splines_offx=0;
my $splines_offy=0;
my $tree;
my $window_zoom;
my $window_width;
my $window_height;
my $window_offx;
my $window_offy;
my $max_window_width=800;
my $max_window_height=730;
my @lastviews    = ();
my @lasttrees    = ();
my @splines      = ();

# options
my $godmode      = 0;
my %info_mode    = ();
my $edit_mode    = "normal";
my $redraw_mode  = 1;
my $markedobj    = "";
my $outputSplinesCmd = "/home/nimbro/checkout/auto/build2/tools/outputspline/outputSpline_METRIC";
my $closest_to_click;
my $abort_mouse_action = 0;
my $minZoomForSpline = 10;
my $minZoomForWPNrs  = 10;

my %operations = (
	"supportpoint-button1"             => "drag-supportpoint",
	"supportpoint-button1-drag"        => "drag-supportpoint",
	"supportpoint-button2"             => "insert-supportpoint",
	"supportpoint-button3"             => "delete-supportpoint",
                                       
	"normal-ctrl-button1-drag"         => "mouse-zoom",
	"normal-ctrl-button3-drag"         => "mouse-pan",
	"laneedit-ctrl-button1-drag"       => "mouse-zoom",
	"laneedit-ctrl-button3-drag"       => "mouse-pan",
                                       
	"laneexit-button1-drag"            => "insert-exit",
	"laneexit-button3-drag"            => "delete-exit",
                                       
	"mark-button1"                     => "mark-point",
	"mark-button2"                     => "mark-lane",
	"mark-button3"                     => "mark-segment",

	"laneedit-ctrl-shift-button1-drag" => "insert-segment",
	"laneedit-ctrl-shift-button3"      => "delete-segment",
	"laneedit-button1-drag"            => "drag-waypoint",
	"laneedit-button1"                 => "drag-waypoint",
	"laneedit-button2"                 => "insert-waypoint",
	"laneedit-button3"                 => "delete-waypoint",
	"laneedit-ctrl-button2"            => "insert-checkpoint",
	"laneedit-ctrl-button3"            => "delete-checkpoint",
	"laneedit-shift-button1"           => "toggle-optional-waypoint",
	"laneedit-shift-button3-drag"      => "drag-splines",
	"laneedit-shift-button3"           => "drag-splines",

	"zoneedit-ctrl-button1"            => "toggle-zonelane",
	"zoneedit-button1-drag"            => "drag-waypoint",
	"zoneedit-button1"                 => "drag-waypoint",
	"zoneedit-button2-drag"            => "insert-zone",
	"zoneedit-button3"                 => "delete-zone",

	"zoneedit-ctrl-button2-drag"       => "insert-spot",
	"zoneedit-ctrl-button3"            => "delete-spot",

	"up"                               => "move-up",
	"down"                             => "move-down",
	"right"                            => "move-right",
	"left"                             => "move-left",
	"kp_add"                           => "zoom-in",
	"ctrl-plus"                        => "zoom-in",
	"ctrl-minus"                       => "zoom-out",
	"kp_subtract"                      => "zoom-out",
                                       
	"space"                            => "toggle-redraw",
	"ctrl-r"                           => "manual-rename",
	"ctrl-b"                           => "backup",
	"ctrl-z"                           => "undo",
	"ctrl-q"                           => "quit",
	"ctrl-g"                           => "whole-image",
	"ctrl-s"                           => "save",
	"ctrl-shift-s"                     => "save-nooptional",
	"ctrl-l"                           => "load",
                                       
	"1"                                => "toggle-normal-mode",
	"2"                                => "toggle-mark-mode",
	"3"                                => "toggle-laneexit-mode",
	"4"                                => "toggle-laneedit-mode",
	"5"                                => "toggle-supportpoint-mode",
	"6"                                => "toggle-zoneedit-mode",
	"7"                                => "toggle-text",
	"8"                                => "toggle-heading",
	"9"                                => "toggle-splines",
	"0"								   => "toggle-godmode",
                                       
	"backspace"                        => "last-view",
	"escape"                           => "abort-mouse-action",
                                       
	"ctrl-h"                           => "help",
);


sub getSplinesMETRIC{ # <<<2
	@splines = ();
	my $sxmin  = min map{$$_[0]} values %centerCalcWP;
	my $sxmax  = max map{$$_[0]} values %centerCalcWP;
	my $symin  = min map{$$_[1]} values %centerCalcWP;
	my $symax  = max map{$$_[1]} values %centerCalcWP;
	my $cx     = 0.5 * ($sxmin + $sxmax);
	my $cy     = 0.5 * ($symin + $symax);
	### metric center is at gps: "$cx, $cy"
	my %gpssupportpoints = map {$_ => $waypoints{$_}} keys %supportpoints;
	my @triangles     = TriangleWrap::getTriangles(\%gpssupportpoints);
	($cx, $cy) = GPST12N::getTransformed(\%gpssupportpoints,\%supportpoints,\@triangles,$cx,$cy);
	### metric center is at img: "$cx, $cy"

	my $metersticklen    = 0.2*$large_height; # pixels
	#### calculating geo coordinates in image
	my @m1               = GPST12N::img2gps(\%supportpoints,\%waypoints,0.5*$large_width,0.1*$large_height);
	my @m2               = GPST12N::img2gps(\%supportpoints,\%waypoints,0.5*$large_width,0.1*$large_height+$metersticklen);
	if($m1[0] == 0 or $m2[0] == 0 or $m1[1] == 0 or $m2[1] == 0){
		warn "problems while determining meter scale for splines: @m1  @m2\n";
		return
	}
	#### Calculating meter stick length
	my $geo              = new Geo::Distance;
	my $metricmetersticklen  = $geo->distance( 'meter', reverse(@m1) => reverse(@m2) );

	### write rndf graph
	my $tmpfile = "tmp_getsplines.rndf";
	save_tree($tmpfile);
	### read splines
	my @last = undef;
	unless(-x $outputSplinesCmd){
		warn "You have to set outputSplinesCmd in the .ini file to an existing executable!";
		return;
	}
	open(S, "$outputSplinesCmd $tmpfile|") or die $!;
	while(<S>){
		chomp;
		if(s/\././g > 2){    # new spline start
			@last = undef;
			next;
		}
		@_ = split /\s+/;
		$_[0]  = $cx + $_[0]*$metersticklen/$metricmetersticklen;
		$_[1]  = $cy - $_[1]*$metersticklen/$metricmetersticklen;
		push @splines, [@last, @_] if(@last);
		@last = @_;
	}
}
sub getSplinesGPS{ # <<<2
	@splines = ();
	my %gpssupportpoints = map {$_ => $waypoints{$_}} keys %supportpoints;
	my @triangles     = TriangleWrap::getTriangles(\%gpssupportpoints);
	### write rndf graph
	my $tmpfile = "tmp_getsplines.rndf";
	save_tree($tmpfile);
	### read splines
	unless(-x $outputSplinesCmd){
		warn "You have to set outputSplinesCmd in the .ini file to an existing executable!";
		return;
	}
	open(S, "$outputSplinesCmd $tmpfile|") or die $!;
	my @pt = ();
	my @last = undef;
	while(<S>){
		chomp;
		if(s/\././g > 2){    # new spline start
			@last = undef;
			next;
		}
		@_  = split /\s+/;
		@pt = GPST12N::getTransformed(\%gpssupportpoints,\%supportpoints,\@triangles,@_);
		push @splines, [@last, @pt] if(@last);
		@last = @pt;
	}
}

sub getImageSize{ # <<<2
	print "determining size of image. press strg-s to save it and speed up start...";
	my $img  = shift;
	my $nipx = qq{'print ((Image_file "$img").rect.width)'};
	my $nipy = qq{'print ((Image_file "$img").rect.height)'};
	my @ret  = ( `/usr/bin/nip2 -e $nipx`,
		`/usr/bin/nip2 -e $nipy`);
	chomp @ret;
	print "$ret[0] x $ret[1]\n";
	return @ret;
}

sub extractDrawables{ # <<<2
	my $tree      = shift;
	@lines        = ();
	@stoppoints   = ();
	@perimeters   = ();
	%exitpoints   = ();
	%checkpoints  = ();
	%waypoints    = ();
	%insertedwp   = ();
	%optionalwp   = ();
	@zonelanes    = ();
	%supportpoints = ();
	my $segments = $tree->segment;
	foreach my $segment (@$segments){ # <<<3
		foreach my $lane (@{$segment->lane}){
			# extract checkpoints, exitpoints, stoppoints
			foreach my $cp (@{$lane->lane_opt}){
				if($cp->rule eq "exit"){
					my $from = $cp->exit_from->id;
					my $to   = $cp->exit_to->id;
					if($from eq $to){
						warn "IGNORING EXIT FROM TO ITSELF";
						next
					}
					push @{$exitpoints{$from}}, $to;
				}elsif($cp->rule eq "checkpoint"){
					my $wp = $cp->id;
					my $nr = $cp->integer;
					if(grep{$_ == $nr}values%checkpoints){
						warn "YOU HAVE CHECKPOINT $nr TWICE";
						next
					}
					$checkpoints{$wp} = $nr;
				}elsif($cp->rule eq "stop"){
					next unless($cp->rule eq "stop");
					my $p = $cp->id;
					push @stoppoints, $p;
				}
			}
			my $isZoneLane = 0;
			foreach my $o (@{$lane->lane_opt}){
				if($o->rule eq "zone_lane"){
					$isZoneLane = 1;
					last;
				}
			}
			# extract waypoints
			my @lanepoints;
			foreach my $wp (@{$lane->waypoint}){
				my $wpid = $wp->id;
				my ($lat,$lon) = $wp->coor;
				if($wp->id1 != $segment->id){
					warn "LANE WAYPOINT $wpid ID DOES NOT MATCH ITS SEGMENT ID " . $segment->id;
				}
				if($wp->id12 != $lane->id){
					warn "LANE WAYPOINT $wpid ID DOES NOT MATCH ITS LANE ID " . $lane->id;
				}
				if(exists $waypoints{$wpid}){
					warn "YOU HAVE TWO DEFINITIONS OF WAYPOINT $wpid!";
				}
				$waypoints{$wpid}     = [ $lat, $lon ];
				$centerCalcWP{$wpid}  = [ $lat, $lon ];
				$insertedwp{$wpid}    = 1 if ($wp->inserted);
				$optionalwp{$wpid}    = 1 if ($wp->optional);
				my @sp = $wp->supportpoint;
				$supportpoints{$wpid} = [@sp] if (scalar @sp);
				push @lanepoints, $wpid;
			}

			# convert waypoints to lines
			foreach my $i (0..$#lanepoints-1){
				my $lp1 = $lanepoints[$i];
				my $lp2 = $lanepoints[$i+1];
				push @lines, [ $lp1 , $lp2 ];
				push @zonelanes, [ $lp1 , $lp2 ], if($isZoneLane);
			}
		}
	}
	my $zones = $tree->zone;
	foreach my $zone (@$zones){ # <<<3
		my @perim;
		foreach my $o (@{$zone->perimeter->peri_opt}){
			if ($o->rule eq "exit"){
				my $from = $o->exit_from->id;
				my $to   = $o->exit_to->id;
				if($from eq $to){
					warn "IGNORING EXIT FROM TO ITSELF";
					next
				}
				push @{$exitpoints{$from}}, $to;
			}
		}
		foreach my $wp (@{$zone->perimeter->waypoint}){
			my $p   = $wp->id;
			my $lat = $wp->latitude;
			my $lon = $wp->longitude;
			if($wp->id1 != $zone->id){
				warn "PERIMETER WAYPOINT $p ID DOES NOT MATCH ITS ZONE ID";
			}
			if($wp->id2 != 0){
				warn "PERIMETER WAYPOINT $p HAS A SECOND ID != 0";
			}
			if(exists $waypoints{$p}){
				warn "YOU HAVE TWO DEFINITIONS OF WAYPOINT $p!";
			}
			$waypoints{$p}        = [$lat,$lon];
			$insertedwp{$p}       = 1 if ($wp->inserted);
			$optionalwp{$p}       = 1 if ($wp->optional);
			$supportpoints{$p} = [$wp->supportpoint] if ($wp->supportpoint);
			push @perim, $p;
		}
		push @perimeters, [ @perim ];
		foreach my $spot (@{$zone->spot}){
			foreach my $opt (@{$spot->spot_opt}){
				if($opt->rule eq "checkpoint"){
					my $wp = $opt->id;
					my $nr = $opt->integer;
					if(grep{$_ == $nr}values%checkpoints){
						warn "YOU HAVE CHECKPOINT $nr TWICE";
						next
					}
					$checkpoints{$wp} = $nr;
				}elsif ($opt->rule eq "spot_width"){
					# ...
				}
			}
			push @lines, [$spot->waypoint->[0]->id, $spot->waypoint->[1]->id];
			foreach my $wp (@{$spot->waypoint}){
				my $p   = $wp->id;
				my $lat = $wp->latitude;
				my $lon = $wp->longitude;
				if($wp->id2 < 1){
					warn "SPOT WAYPOINT $p HAS A SECOND ID < 1!";
				}
				if(exists $waypoints{$p}){
					warn "YOU HAVE TWO DEFINITIONS OF WAYPOINT $p!";
				}
				$waypoints{$p} = [$lat,$lon];
			}
		}
	}
	my $extsp = $tree->{'extsupportpoint(s)'};
	if(scalar keys %supportpoints < 5 and ref $extsp and scalar(@$extsp)){
		### Number of ext supportpoints: scalar @$extsp
		### Less than 5 supportpoints in $rndffile, but there are supportpoints after the end of the file
		foreach my $esp (@$extsp){ # <<<3
			$supportpoints{$esp->string} = [$esp->xcoor, $esp->ycoor];
			$waypoints{$esp->string} = [$esp->latitude, $esp->longitude];
		}
	}elsif(scalar keys %supportpoints >= 5){
		my  @extsp;
		while(my ($p,$ic) = each(%supportpoints)){
			my $gc = $waypoints{$p};
			next unless $gc;
			push @extsp, new RNDF::extsupportpoint(string => "ext$p", latitude => $gc->[0], longitude => $gc->[1], xcoor => $ic->[0], ycoor => $ic->[1] );
		}
		$tree->{'extsupportpoint(s)'} = \@extsp;
	}else{
		print "warning: not enough supportpoints in $rndffile, no alignment with image possible.\n";
	}
}

sub dist{  # <<<2
	return sqrt(dist2(@_));
}
sub dist2{ # <<<2
	my ($a,$b) = @_;
	return ($$b[0]-$$a[0])**2 + ($$b[1]-$$a[1])**2
}
sub cleanup_imagecache{
	my @cache = </tmp/rndf-*.png>;
	if (scalar @cache > 40){
		### cleanup cache...
		@cache = sort{stat($a)->atime <=> stat($b)->atime}@cache;
		File::Remove::remove(shift @cache) while (scalar @cache > 20);
	}
}
sub makeSubImg{ # <<<2
	cleanup_imagecache();

	my $largeimg = shift;
	my $code     = md5_hex($largeimg);
	my ($width,$height,$xoff,$yoff,$zoom) = @_;
	my $smallimg = "/tmp/rndf-$code-$width-$height-$xoff-$yoff-$zoom.png";
	if (-e $smallimg){
		### using cached image...
		return $smallimg;
	}
	my $zoomed_w = $width*$zoom;
	my $zoomed_h = $height*$zoom;
	my $z = 1.0/$zoom;
	my $nip = qq{extract_area $xoff $yoff $width $height (resize $z $z 0 (Image_file "$largeimg"));};
	$nip =~ s/,/./g;
	### $nip
	### Preparing Sub-Image, this will take some time, but would you want the whole image in your RAM instead? ";
	system("/usr/bin/nip2","-e",$nip,"-o",$smallimg); 
	$smallimg;
}

sub run { # <<<2
	$largefile = shift;
	$tree      = shift;
	$rndffile  = shift;
	unless(load_settings()){
		($large_width, $large_height) = getImageSize($largefile);
		$window_width  = $max_window_width;
		$window_height = int($large_height / $large_width * $window_width);
		$window_height = min($window_height,$max_window_height);
		$window_offx   = 0;
		$window_offy   = 0;
		$window_zoom   = 1;
	}

	$win = Gtk2::CV::ImageWindow->new;
	$win->set_title("RNDF Editor");
	$win->signal_connect("button_press_event"   => sub { button_press(1, $_[1]) });
	$win->signal_connect("button_release_event" => sub { button_press(0, $_[1]) });
	$win->signal_connect("key_press_event"      => sub { key_press($_[1]->keyval, $_[1]->state) });
	$win->show_all;
	redraw(1,1);
	### GUI initialized.
	Gtk2->main;
}

sub is_marked{ # <<<2
	my $id = shift;
	return 0 unless $markedobj;
	return 1 if ($id =~ /^$markedobj\b/);
	return 0;
}

my $bgimg;
my $gimg;
my $cimg;
sub redraw{ # <<<2
	return unless $redraw_mode;
	my $update_bg      = shift;
	my $update_graph   = shift;
	my $update_spline  = 1;
	if(@_){
		$update_spline  = shift;
	}
	
	#my $progress = new Gtk2::CV::Progress title => "Redrawing", work => 5;
	if($update_bg){ # <<<3
		### load background image
		#800 x 747 +4000+3735 from 16015 x 14942 z=10
		#800 x 747 +4000+3735 from 16015 x 14942 z=20
		print "SubImage Extract $window_width x $window_height +$window_offx+$window_offy from $large_width x $large_height z=$window_zoom\n\n";
		my $bgfile = makeSubImg($largefile,$window_width,$window_height,int($window_offx/$window_zoom),int($window_offy/$window_zoom),$window_zoom);
		$bgimg  = GD::Simple->newFromPng($bgfile);
	}

	if($update_graph){ # <<<3
		# graph image
		### extracting data from tree...
		extractDrawables($tree);
		### creating image for graph...
		$gimg  = GD::Simple->new($bgimg->width, $bgimg->height);
		my $white = $gimg->colorClosest(255,255,255);
		$gimg->transparent($white);
		### scaling waypoints...
		%wptransformed = GPST12N::getTransformedWaypoints(\%supportpoints,\%waypoints,$window_zoom,$window_offx,$window_offy,$large_width);

		#### splines
		if($info_mode{splines} and $window_zoom <= $minZoomForSpline){
			if($update_spline){
				getSplinesMETRIC() if($info_mode{splines} eq 1);
			}
			$gimg->fgcolor('white');
			$gimg->penSize(1);
			SLANE: foreach my $l (@splines){
				map{next SLANE unless($_)} @$l;
				my ($x1, $y1, $x2, $y2) = @$l;
				($x1, $y1) = large2win($splines_offx+$x1,$splines_offy+ $y1);
				($x2, $y2) = large2win($splines_offx+$x2,$splines_offy+ $y2);
				next unless ($gimg->boundsSafe($x1, $y1) or $gimg->boundsSafe($x2, $y2));
				$gimg->moveTo($x1, $y1);
				$gimg->lineTo($x2, $y2);
			}
		}

		### [<now>] drawing graph... with zoom: $window_zoom
		#### Perimeters...
		PERI: foreach my $peri (@perimeters){
			my $poly = new GD::Polygon;
			foreach (@$peri){
			 	my $p = $wptransformed{$_};
				warn "undefined peripoint $_" unless $p;
				next PERI unless $gimg->boundsSafe(@$p);
				$poly->addPt(@$p);
			};
			$gimg->fgcolor('orange');
			$gimg->bgcolor(undef);
			$gimg->polygon($poly);
		}
		#### Supportpoints...
		$gimg->fgcolor('yellow');
		$gimg->bgcolor(undef);
		foreach my $p (keys %supportpoints){
			next unless $gimg->boundsSafe(@{$wptransformed{$p}});
			$gimg->moveTo(@{$wptransformed{$p}});
			$gimg->ellipse(15,15);
		}
		#### Checkpoints...
		foreach my $ew (keys %checkpoints){
		    my $w = $wptransformed{$ew};
			warn "invalid checkpoint '$ew'" unless $w;
		    next unless $gimg->boundsSafe(@$w);
			$gimg->fgcolor('green');
			$gimg->bgcolor('green');
			$gimg->rectangle($$w[0]-3,$$w[1]-3,$$w[0]+3,$$w[1]+3);
			if ($info_mode{textual}){
				my $s = $checkpoints{$ew};
				my ($dx,$dy) = $gimg->stringBounds($s);
				my ($x,$y)   = ($$w[0]-10.0,$$w[1]-15.0);
				$gimg->rectangle($x-1,$y-1,$x+$dx+1,$y+$dy+1);
				$gimg->moveTo($x+1,$y+$dy);
				$gimg->fgcolor('yellow');
				$gimg->string($checkpoints{$ew});
			}
		}
		#### Waypoints
		$gimg->bgcolor(undef);
		$gimg->fgcolor('blue');
		while(my ($v,$w) = each(%wptransformed)){
		    next unless $gimg->boundsSafe(@$w);
			if (exists $optionalwp{$v}){
				$gimg->fgcolor('red');
				$gimg->bgcolor('blue');
			}elsif (exists $insertedwp{$v}){
				$gimg->fgcolor('blue');
				$gimg->bgcolor('red');
			}else{
				$gimg->fgcolor('blue');
				$gimg->bgcolor(undef);
			}
			$gimg->rectangle($$w[0]-2.0,$$w[1]-2.0,$$w[0]+2.0,$$w[1]+2.0);
			if ($info_mode{textual} and $window_zoom <= $minZoomForWPNrs){
				$gimg->moveTo($$w[0]-10.0,$$w[1]+15.0);
				$gimg->string($v);
			}
		}
		#### Stoppoints
		$gimg->fgcolor('white');
		$gimg->bgcolor('red');
		foreach my $w (@stoppoints){
			$w = $wptransformed{$w};
			warn "undef stoppoint" unless $w;
			next unless $gimg->boundsSafe(@$w);
			# $gimg->moveTo($$w[0]+5,$$w[1]+5);
			# $gimg->ellipse(7,7);
		}
		#### Lanes
		foreach my $l (@lines){
			my $lp1 = $wptransformed{$$l[0]};
			my $lp2 = $wptransformed{$$l[1]};
			if (is_marked($$l[0]) or is_marked($$l[1])){
				$gimg->fgcolor('red');
			}
			else {
				$gimg->fgcolor('blue');
			}
		    next unless ($gimg->boundsSafe(@$lp1) or $gimg->boundsSafe(@$lp2));
			$gimg->moveTo(@$lp1);
			if ($info_mode{heading}){     # show thick lines half way
				$gimg->penSize(3);
				my @lp1a = (
					$$lp1[0] + 0.5*($$lp2[0]-$$lp1[0]),  
					$$lp1[1] + 0.5*($$lp2[1]-$$lp1[1]));
				$gimg->lineTo(@lp1a);
			}
			$gimg->penSize(1);
			$gimg->lineTo(@$lp2);
		}
		#### zoneLanes
		foreach my $l (@zonelanes){
			my $lp1 = $wptransformed{$$l[0]};
			my $lp2 = $wptransformed{$$l[1]};
			if (is_marked($$l[0]) or is_marked($$l[1])){
				$gimg->fgcolor('red');
			}
			else {
				$gimg->fgcolor('black');
			}
		    next unless ($gimg->boundsSafe(@$lp1) or $gimg->boundsSafe(@$lp2));
			$gimg->moveTo(@$lp1);
			if ($info_mode{heading}){     # show thick lines half way
				$gimg->penSize(3);
				my @lp1a = (
					$$lp1[0] + 0.5*($$lp2[0]-$$lp1[0]),  
					$$lp1[1] + 0.5*($$lp2[1]-$$lp1[1]));
				$gimg->lineTo(@lp1a);
			}
			$gimg->penSize(1);
			$gimg->lineTo(@$lp2);
		}
		#### Exits
		$gimg->fgcolor('blue');
		$gimg->bgcolor('yellow');
		foreach my $ew (keys %exitpoints){
			my $w = $wptransformed{$ew};
			warn "undef exitpoint $ew" unless $w;
		    next unless $gimg->boundsSafe(@$w);
			$gimg->fgcolor('blue');
			$gimg->rectangle($$w[0]-2,$$w[1]-2,$$w[0]+2,$$w[1]+2);
			foreach my $ev (@{$exitpoints{$ew}}){
				my $v = $wptransformed{$ev};
				warn "undef entrypoint $ev" unless $v;
				next unless $gimg->boundsSafe(@$v);
				$gimg->fgcolor('yellow');
				$gimg->moveTo(@$w);
				$gimg->lineTo(@$v);
			}
		}
		if($godmode){
			#### Godmode
			$gimg->fgcolor('red');
			$gimg->bgcolor(undef);
			$gimg->penSize(10);
			$gimg->rectangle(10,10,$window_width-10, $window_height-10);
		}
		if(0 and scalar keys %supportpoints > 4){
			#### Masstab
			my $meter_stick_len = $window_height/2;
			my @m1 = GPST12N::img2gps(\%supportpoints,\%waypoints,win2largeimg(0,0));
			my @m2 = GPST12N::img2gps(\%supportpoints,\%waypoints,win2largeimg(0,$meter_stick_len));
			my $geo       = new Geo::Distance;
			my $distance  = exec '$geo->distance( "meter", reverse(@m1) => reverse(@m2) )';
			if($distance){
				my $dcut;
				if ($distance>100){
					$dcut     = int($distance/100)*100;
				}elsif ($distance>10){
					$dcut     = int($distance/10)*10;
				}elsif ($distance>1){
					$dcut     = int($distance/1)*1;
				}else {
					$dcut     = int($distance*10)/10;
				}
				my $h2       = $dcut/$distance*$meter_stick_len;
				$gimg->penSize(2);
				$gimg->fgcolor('white');
				$gimg->bgcolor('white');
				my ($px,$py) = (20,60);
				$gimg->moveTo($px,$py);
				$gimg->rectangle($px-5,$py-2,$px+5,$py+2);
				$py+=$h2;
				$gimg->lineTo($px,$py);
				$gimg->rectangle($px-5,$py-2,$px+5,$py+2);
				$py+=20;
				$gimg->penSize(1);
				$gimg->moveTo($px,$py);
				my $s = "$dcut m";
				my @sb = $gimg->stringBounds($s);
				$gimg->rectangle($px-5-2,$py,$px+$sb[0]-5+2,$py-$sb[1]);
				$gimg->fgcolor('black');
				$gimg->moveTo($px-5,$py);
				$gimg->string("$dcut m");
				$gimg->fgcolor('white');
				$gimg->moveTo($px-4,$py+1);
				$gimg->string("$dcut m");
			}
		}

		### showing additional gps points
		my $agp = "offsets_ao_0.00_0.00.dat";
		if(-e $agp){
			my %wp2 = %waypoints;
			open FH, "<$agp" or $agp;
			my %gpssupportpoints = map {$_ => $waypoints{$_}} keys %supportpoints;
			my @triangles        = TriangleWrap::getTriangles(\%gpssupportpoints);
			my $cnt = 0;
			$gimg->penSize(1);
			while(<FH>){
				@_ = split/\s+/;
				my ($g1, $g2) = ($_[21],$_[20]);
				my ($u1, $u2) = ($_[23],$_[22]);
				my ($t1, $t2) = ($_[6],$_[7]);
				my @pt = GPST12N::getTransformed(\%gpssupportpoints,\%supportpoints,\@triangles,$g1,$g2);
				my @ut = GPST12N::getTransformed(\%gpssupportpoints,\%supportpoints,\@triangles,$u1,$u2);
				#print "transformed $g1, $g2 to $pt[0], $pt[1]\n";
				@pt = large2win(@pt);
				@ut = large2win(@ut);
				if($t1 == 0 or $t2 == 0) {
					$gimg->bgcolor('white'); $gimg->fgcolor('green');
					next;
				}else{
					$gimg->bgcolor('white'); $gimg->fgcolor('red');
				}
				$gimg->bgcolor('white'); $gimg->fgcolor('green');
				$gimg->moveTo($pt[0], $pt[1]);
				$gimg->ellipse(3,3);
				$gimg->bgcolor('white'); $gimg->fgcolor('red');
				$gimg->moveTo($ut[0], $ut[1]);
				$gimg->ellipse(3,3);
			}
			close FH;
		}

		### combining images...
		$cimg = $bgimg->clone;
		$cimg->copy($gimg->gd,0,0,0,0,$bgimg->width,$bgimg->height);
	}
	#### Generating Help message
	if ($info_mode{keybind}){
		my $himg  = GD::Simple->new($bgimg->width-40, $bgimg->height-40);
		$himg->bgcolor('white');
		$himg->fgcolor('black');
		$himg->rectangle(0,0,$himg->width,$himg->height);
		my ($px,$py) = (10,10);
		$himg->moveTo($px,$py);
		my $s = "Available Mouse Actions in $edit_mode-mode:";
		my ($dx,$dy) = $gimg->stringBounds($s);
		$py += $dy;
		$himg->moveTo($px,$py);
		$himg->string($s);
		$py += 5;
		map{my $s = sprintf "%-30s %-30s",$_,$operations{$_};
			$s =~ s/^$edit_mode-//;
			($dx,$dy) = $gimg->stringBounds($s);
			$py += $dy;
			$himg->moveTo($px,$py);
			$himg->string($s);
		}grep{/^$edit_mode-/}sort grep{/button\d/}keys(%operations);
		$py += 20;
		$s = "Available Keyboard Actions:";
		($dx,$dy) = $gimg->stringBounds($s);
		$py += $dy;
		$himg->moveTo($px,$py);
		$himg->string($s);
		$py += 5;
		map{
			my $s = sprintf "%-30s %-30s",$_,$operations{$_};
			my ($dx,$dy) = $gimg->stringBounds($s);
			$py += $dy;
			$himg->moveTo($px,$py);
			$himg->string($s);
		}sort grep{!/button\d/}keys(%operations);
		$cimg->copy($himg->gd,20,20,0,0,$himg->width,$himg->height);
	}else{
		my $himg  = GD::Simple->new(200,50);
		my $white = $himg->colorClosest(255,255,255);
		$himg->transparent($white);
		$himg->bgcolor('white');
		$himg->fgcolor('black');
		my ($px,$py) = (5,5);
		my $s = "Mode: $edit_mode. Help: Ctrl-H";
		my ($dx,$dy) = $himg->stringBounds($s);
		$himg->rectangle($px,$py,$px+$dx+10,$py+$dy+10);
		$px+= 5;
		$py+= $dy+5;
		$himg->moveTo($px,$py);
		$himg->string($s);
		$cimg->copy($himg->gd,10,10,0,0,$himg->width,$himg->height);
	}


	my $img_data = $cimg->ppm;
	$img_data =~ s/^P6\s+\d+\s+\d+\s+\d+\n//;

	#### copying to pixbuf...
	my ($w,$h) = ($cimg->width,$cimg->height);
	my $pbuf = Gtk2::Gdk::Pixbuf->new_from_data($img_data,"rgb",0,8,$w,$h,3*$w);
	$win->set_image($pbuf);
	### done drawing.
}

sub load_settings{ # <<<2
	### loading settings...
	my $code = md5_hex($largefile);
	return 0 unless($cfg->SectionExists($code));
	$large_width   = $cfg->val($code, "large_width") ;
	$large_height  = $cfg->val($code, "large_height") ;
	$window_width  = $cfg->val($code, "window_width") ;
	$window_zoom   = $cfg->val($code, "window_zoom") ;
	$window_offx   = $cfg->val($code, "window_offx") ;
	$window_offy   = $cfg->val($code, "window_offy") ;
	my $tmp = $cfg->val("Default", "outputSplinesCmd");
	$outputSplinesCmd = $tmp if($tmp);

	$tmp = $cfg->val("Default", "max_window_width");
	$max_window_width = $tmp if($tmp);
	$tmp = $cfg->val("Default", "max_window_height");
	$max_window_height = $tmp if($tmp);

	$window_height = int($large_height / $large_width * $window_width);
	$window_height = min($window_height, $max_window_height);

	foreach my $p ($cfg->Parameters("GUI_Operations")){
		$operations{$p} = $cfg->val("GUI_Operations",$p);
	}
	return 1;
}
sub undo { # <<<2
	$tree = shift @lasttrees || $tree;
}
sub save_tree_for_undo { # <<<2
	unshift @lasttrees, clone($tree);
	#unshift @lasttrees, make_cow_ref($tree);  # that aint workin
	pop @lasttrees if (scalar @lasttrees>25);
}
sub save_view { # <<<2
	### Saving last view
	unshift @lastviews, qq{
		\$window_offx   = $window_offx;
		\$window_offy   = $window_offy;
		\$window_width  = $window_width;
		\$window_height = $window_height;
		\$window_zoom   = $window_zoom;
	};
}
sub save_settings{ # <<<2
	### saving your settings...
	my $code  = md5_hex($largefile);
	$cfg->newval("Default",$code,$largefile);
	$cfg->newval("Default","outputSplinesCmd", $outputSplinesCmd);
	$cfg->newval("Default","max_window_width", $max_window_width);
	$cfg->newval("Default","max_window_height", $max_window_height);
	$cfg->AddSection($code);
	$cfg->AddSection("GUI_Operations");
	$cfg->newval($code,"large_width" , $large_width);
	$cfg->newval($code,"large_height" ,$large_height);
	$cfg->newval($code,"window_zoom" , $window_zoom);
	$cfg->newval($code,"window_width" , $window_width);
	$cfg->newval($code,"window_offx" , $window_offx);
	$cfg->newval($code,"window_offy" , $window_offy);
	while (my ($k,$v) = each(%operations)){
		$cfg->newval("GUI_Operations",$k,$v);
	}
	$cfg->RewriteConfig;
}

sub save_tree{ # <<<2
	my $fn = shift;
	$fn = $rndffile unless ($fn);
	### Saving rndf-tree to: $fn
	open O, ">$fn" or die $!;
	print O $tree->rndf;
	close O;
	### Saving done.
}
sub save_tree_nooptional{ # <<<2
	### Saving tree without optional components
	my $cpy = clone($tree);
	$cpy->removeWaypointsWithFlag("optional");
	open O, ">$rndffile-nooptional" or die $!;
	print O $cpy->rndf;
	close O;
	### Saving tree done.
}

sub key_press{ # <<<2
	my ($key,$state) = @_; 
	my $zoomfact  = 2.0;
	my $transfact = 0.25;

	my $ctrl    = ($state * "control-mask")?"ctrl-":"";
	my $shift   = ($state * "shift-mask")?"shift-":"";
	my %rkeysyms= reverse %Gtk2::Gdk::Keysyms;
	my $keyname = $rkeysyms{$key};
	my $action  = lc "$ctrl$shift$keyname";
	my $cmd     = $operations{$action};
	if (!$cmd){
		if ($action ne "alt_l" and $action ne "control_l"){
			print "no cmd defined for action $action\n";
		}
		return 1;
	}

	### Selected command: $cmd
	if($cmd eq "quit"){
		exit(0);
	}elsif($cmd eq "undo"){
		$tree = shift @lasttrees;
		redraw(0,1,1);
	}elsif($cmd eq "toggle-redraw"){
		$redraw_mode ^= 1;
		redraw(0,1,0);
		return 1;
	}elsif($cmd eq "toggle-godmode"){
		$godmode ^= 1;
		redraw(0,1,0);
		return 1;
	}elsif($cmd eq "toggle-splines"){
		$info_mode{splines}++;
		$info_mode{splines}%=3;
		redraw(0,1,1);
		return 1;
	}elsif($cmd eq "toggle-heading"){
		$info_mode{heading} ^= 1;
		redraw(0,1,0);
		return 1;
	}elsif($cmd eq "toggle-zoneedit-mode"){
		$edit_mode = "zoneedit";
		redraw(0,0,0);
		return 1;
	}elsif($cmd eq "toggle-supportpoint-mode"){
		$edit_mode = "supportpoint";
		redraw(0,0,0);
		return 1;
	}elsif($cmd eq "toggle-laneedit-mode"){
		$edit_mode = "laneedit";
		redraw(0,0,0);
		return 1;
	}elsif($cmd eq "toggle-laneexit-mode"){
		$edit_mode = "laneexit";
		redraw(0,0,0);
		return 1;
	}elsif($cmd eq "toggle-normal-mode"){
		$edit_mode = "normal";
		redraw(0,0,0);
		return 1;
	}elsif($cmd eq "toggle-mark-mode"){
		$markedobj = "";
		$edit_mode = "mark";
		redraw(0,0,0);
		return 1;
	}elsif($cmd eq "toggle-text"){
		$info_mode{textual} ^= 1;
		redraw(0,1,0);
		return 1;
	}elsif($cmd eq "abort-mouse-action"){
		print "Aborting mouse action!\n";
		$abort_mouse_action = 1;
		return 1;
	}elsif($cmd eq "last-view"){
		if (scalar @lastviews){
			eval shift @lastviews 
		}else{
			warn "already at 1st view, can't go back\n";
			return 1
		}
	}elsif($cmd eq "help"){
		$info_mode{keybind} ^= 1;
		redraw(0,1,0);
		return 1;
	}elsif($cmd eq "whole-image"){
		save_view();
		$window_offx = 0;
		$window_offy = 0;
		$window_zoom = $large_width/$window_width; 
		$window_zoom = int($window_zoom) unless ($window_zoom<1);
	}elsif($cmd eq "zoom-in"){
		save_view();
		my $A = int($window_width *$window_zoom*(1-1/$zoomfact)/2);
		my $B = int($window_height*$window_zoom*(1-1/$zoomfact)/2);
		$window_offx += $A;
		$window_offy += $B;
		$window_zoom /= $zoomfact;
		$window_zoom  = int($window_zoom) unless($window_zoom<1);
	}elsif($cmd eq "zoom-out"){
		save_view();
		my $A = int($window_width *$window_zoom * ($zoomfact-1)/2);
		my $B = int($window_height*$window_zoom * ($zoomfact-1)/2);
		$window_offx -= $A;
		$window_offy -= $B;
		$window_offx  = min($large_width -$window_width *$window_zoom*$zoomfact,$window_offx);
		$window_offy  = min($large_height-$window_height*$window_zoom*$zoomfact,$window_offy);
		$window_offx  = max(0,$window_offx);
		$window_offy  = max(0,$window_offy);
		$window_zoom *= $zoomfact;
		$window_zoom  = int($window_zoom) unless($window_zoom<1);
		$window_zoom  = max(int($window_width/$large_width),$window_zoom);
		$window_zoom  = int($large_width/$window_width) if ($window_width*$window_zoom>$large_width);
	}elsif($cmd eq "move-right"){
		save_view();
		$window_offx += int($transfact * $window_width  * $window_zoom);
		$window_offx  = min($large_width-$window_width  * $window_zoom,$window_offx);
	}elsif($cmd eq "move-left"){
		save_view();
		$window_offx -= int($transfact * $window_width  * $window_zoom);
		$window_offx  = max(0,$window_offx);
	}elsif($cmd eq "move-up"){
		save_view();
		$window_offy -= int($transfact * $window_height * $window_zoom);
		$window_offy  = max(0,$window_offy);
	}elsif($cmd eq "move-down"){
		save_view();
		$window_offy += int($transfact * $window_height * $window_zoom);
		$window_offy  = min($large_height-$window_height*$window_zoom,$window_offy);
	}elsif($cmd eq "save-nooptional"){
		save_tree_nooptional();
		return 1;
	}elsif($cmd eq "manual-rename"){
		print "write: from to\n";
		@_ = split /\s+/,<STDIN>;
		$tree->rename({$_[0] => $_[1]});
	}elsif($cmd eq "backup"){
		my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
		$year += 1900;
		save_tree($rndffile . "-$year-$mon-$mday-$hour$min");
		return 1;
	}elsif($cmd eq "save"){
		save_settings();
		save_tree();
		return 1;
	}elsif($cmd eq "load"){
		load_settings();
		$tree   = RndfParser::get($rndffile)->{startrule};
		redraw(1,1,1);
		return 1;
	}else{
		warn "command $cmd ($action) not implemented\n";
		return 0
	}
	redraw(1,1,0);
	return 1
}

sub large2win { # <<<2
		$_[0] = int($_[0]/$window_zoom - $window_offx/$window_zoom +0.5);
		$_[1] = int($_[1]/$window_zoom - $window_offy/$window_zoom +0.5);
		return (@_)
}
sub win2largeimg { # <<<2
	my ($x,$y) = @_;
	# transform to largeimage coordinates
	$x = $x*$window_zoom + $window_offx;
	$y = $y*$window_zoom + $window_offy;

	return ($x,$y);
}

sub renameStuffInTree {
	my $rename = shift;
	#foreach my $sp (keys %supportpoints){    # important: rename supportpoints
		#if ($$rename{$sp}){
			#my $a = $supportpoints{$sp};
			#delete $supportpoints{$sp};
			#$supportpoints{$$rename{$sp}} = $a;
		#}else{
			#foreach my $sp (keys %supportpoints){
				#foreach my $n (keys %$rename){
					#if ($sp =~ /^$n\./){
						#my $a = $supportpoints{$sp};
						#delete $supportpoints{$sp};
						#my $e = $$rename{$n};
						#$sp =~ s/^$n\./$e./;
						#$supportpoints{$sp} = $a;
					#}
				#}
			#}
		#}
	#}
}

my @start_click;
sub button_press { # <<<2
	my ($press, $event ) = @_;
	my $ctrl   = ($event->state * "control-mask")?"ctrl-":"";
	my $shift  = ($event->state * "shift-mask")  ?"shift-":"";
	my $drag   = (!$press and (abs($event->x-$start_click[0])>20 or abs($event->y-$start_click[1]) > 20))?"-drag":"";
	my $button = "button".($event->button);

	my $action = lc "$edit_mode-$ctrl$shift$button$drag";


	if($press){ 
		@start_click = ($event->x,$event->y);
		$closest_to_click = reduce { dist2($wptransformed{$a},\@start_click) < dist2($wptransformed{$b},\@start_click) ? $a : $b } (keys %wptransformed);
		### Closest point to your click: $closest_to_click
		@_ = GPST12N::img2gps(\%supportpoints,\%waypoints, win2largeimg(@start_click));
		print "you clicked on GPS @_\n";
	}else{
		if($abort_mouse_action){
			print "Mouse action aborted!\n";
			$abort_mouse_action = 0;
			return;
		}
		my $cmd = $operations{$action};
		if (!$cmd){
			warn "No command defined for action $action\n";
			return 1;
		}
		### Selected command: $cmd
		if ($cmd eq "delete-supportpoint"){
			save_tree_for_undo();
			my $p = [$event->x,$event->y];
			$closest_to_click = reduce { dist2($wptransformed{$a},$p) < dist2($wptransformed{$b},$p) ? $a : $b } (keys %supportpoints);
			return 1 if(dist($wptransformed{$closest_to_click},$p)>30); # at least _try_ to hit it ;)
			delete $supportpoints{$closest_to_click};
			$p = $tree->getWaypointByWaypoint($closest_to_click);
			if ($p->supportpoint){
				$p->supportpoint(undef);
			}
			print "deleted $closest_to_click from set of support points\n";
			redraw(0,1,1);
		}elsif($cmd eq "insert-supportpoint"){
			#$supportpoints{$closest_to_click} = [win2largeimg(@{$wptransformed{$closest_to_click}})];
			save_tree_for_undo();
			my $p = $tree->getWaypointByWaypoint($closest_to_click);
			if ($p->inserted){
				warn "cannot make inserted point a waypoint";
				return 1;
			}
			$p->supportpoint(win2largeimg(@{$wptransformed{$closest_to_click}}));
			redraw(0,1,1);
		}elsif($cmd eq "drag-supportpoint"){
			save_tree_for_undo();
			my @newloc = ($event->x, $event->y);
			@newloc    = win2largeimg(@newloc);
			my $p = $tree->getWaypointByWaypoint($closest_to_click);
			unless($p->supportpoint){
				warn "This is not a supportpoint!";
				return 1;
			}
			$p->supportpoint(@newloc);
			redraw(0,1,1);
		}elsif($cmd eq "drag-waypoint"){
			save_tree_for_undo();
			@_ = ($event->x, $event->y);
			@_ = win2largeimg(@_);
			my ($ex,$ey) = win2largeimg($event->x,$event->y);
			my ($gx,$gy) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex,$ey);
			if ($supportpoints{$closest_to_click}){
				warn "I will not move supportpoints.\n";
			}else{
				$tree->moveWaypoint($closest_to_click,$gx,$gy, $godmode);
				redraw(0,1,1);
			}
		}elsif($cmd eq "delete-waypoint"){
			save_tree_for_undo();
			my ($ex,$ey) = win2largeimg($event->x,$event->y);
			my ($gx,$gy) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex,$ey);
			my ($hx,$hy) = @{$waypoints{$closest_to_click}};
			my $rename = $tree->delWayPoint($closest_to_click, $godmode);
			renameStuffInTree($rename);
			redraw(0,1,1);
		}elsif($cmd eq "mark-point"){
			$markedobj = $closest_to_click;
			redraw(0,1,0);
		}elsif($cmd eq "mark-lane"){
			$markedobj = $closest_to_click;
			$markedobj =~ s/\.\d+$//;
			redraw(0,1,0);
		}elsif($cmd eq "mark-segment"){
			$markedobj = $closest_to_click;
			$markedobj =~ s/\.\d+\.\d+$//;
			redraw(0,1,0);
		}elsif($cmd eq "delete-exit"){
			save_tree_for_undo();
			my @p    = ($event->x,$event->y);
			my $wp2  = reduce { dist2($wptransformed{$a},\@p) < dist2($wptransformed{$b},\@p) ? $a : $b } (keys %wptransformed);
			if (!$tree->isZone($closest_to_click)){
				my $lane = $tree->getLaneByWaypoint($closest_to_click);
				print "deleting exit from $closest_to_click to $wp2\n";
				my @newlopt = grep{
					my $a = $_;
					if($a->rule              ne "exit"
						or $a->exit_from->id ne $closest_to_click
						or $a->exit_to->id   ne $wp2){1}
					else{0}
				}@{$lane->lane_opt};
				$lane->{'lane_opt(s?)'} = \@newlopt;
			}else{
				my $zone = $tree->getZoneByWaypoint($closest_to_click);
				print "deleting exit from $closest_to_click to $wp2\n";
				my @newlopt = grep{
					my $a = $_;
					if($a->rule              ne "exit"
						or $a->exit_from->id ne $closest_to_click
						or $a->exit_to->id   ne $wp2){1}
					else{0}
				}@{$zone->perimeter->peri_opt};
				$zone->perimeter->{'peri_opt(s?)'} = \@newlopt;
			}
			redraw(0,1,1);
		}elsif($cmd eq "insert-exit"){
			save_tree_for_undo();
			my @p    = ($event->x,$event->y);
			my $wp2  = reduce { dist2($wptransformed{$a},\@p) < dist2($wptransformed{$b},\@p) ? $a : $b } (keys %wptransformed);
			if ($tree->isZone($closest_to_click)){
				my $zone = $tree->getZoneByWaypoint($closest_to_click);
				my $ef   = new RNDF::exit_from(id => $closest_to_click);
				my $et   = new RNDF::exit_to  (id => $wp2);
				if($closest_to_click eq $wp2){
					warn "Exit from to same point\n";
					return 1;
				}
				push @{$zone->perimeter->peri_opt},new RNDF::peri_opt(__STRING1__=>"exit",exit_from =>$ef, exit_to => $et, comment => "Inserted");
			}else{
				my $lane = $tree->getLaneByWaypoint($closest_to_click);
				my $ef   = new RNDF::exit_from(id => $closest_to_click);
				my $et   = new RNDF::exit_to  (id => $wp2);
				if($closest_to_click eq $wp2){
					warn "Exit from to same point\n";
					return 1;
				}
				push @{$lane->lane_opt},new RNDF::lane_opt(__STRING1__=>"exit",exit_from =>$ef, exit_to => $et, comment => "Inserted");
			}
			redraw(0,1,1);
		}elsif($cmd eq "insert-spot"){
			save_tree_for_undo();
			my ($ex1,$ey1) = win2largeimg(@start_click);
			my ($ex2,$ey2) = win2largeimg($event->x,$event->y);
			my ($gx1,$gy1) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex1,$ey1);
			my ($gx2,$gy2) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex2,$ey2);
			$tree->addSpot($gx1,$gy1,$gx2,$gy2,\%checkpoints);
			redraw(0,1,0);
		}elsif($cmd eq "delete-spot"){
			save_tree_for_undo();
			$tree->delZoneByWaypoint($closest_to_click);
			redraw(0,1,0);
		}elsif($cmd eq "delete-zone"){
			save_tree_for_undo();
			$tree->delZoneByWaypoint($closest_to_click);
			redraw(0,1,0);
		}elsif($cmd eq "insert-zone"){
			save_tree_for_undo();
			my ($ex1,$ey1) = win2largeimg(@start_click);
			my ($ex2,$ey2) = win2largeimg($event->x,$event->y);
			my ($gx1,$gy1) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex1,$ey1);
			my ($gx2,$gy2) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex2,$ey2);
			$tree->addZone($gx1,$gy1,$gx2,$gy2);
			redraw(0,1,0);
		}elsif($cmd eq "insert-segment"){
			save_tree_for_undo();
			my ($ex1,$ey1) = win2largeimg(@start_click);
			my ($ex2,$ey2) = win2largeimg($event->x,$event->y);
			my ($gx1,$gy1) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex1,$ey1);
			my ($gx2,$gy2) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex2,$ey2);
			$tree->addSegment($gx1,$gy1,$gx2,$gy2);
			redraw(0,1,1);
		}elsif($cmd eq "delete-segment"){
			save_tree_for_undo();
			redraw(0,1,1) if($tree->delSegmentByWaypoint($closest_to_click,$godmode));
		}elsif($cmd eq "delete-checkpoint"){
			save_tree_for_undo();
			$tree->delCheckPoint($closest_to_click);
			redraw(0,1,0);
		}elsif($cmd eq "insert-checkpoint"){
			save_tree_for_undo();
			$tree->addCheckPoint($closest_to_click,\%checkpoints);
			redraw(0,1,0);
		}elsif($cmd eq "toggle-zonelane"){
			save_tree_for_undo();
			my $lane = $tree->getLaneByWaypoint($closest_to_click);
			my @wozoneopt = grep {$_->rule ne "zone_lane"} @{$lane->lane_opt};
			if(scalar(@wozoneopt) != scalar(@{$lane->lane_opt})){
				$lane->{'lane_opt(s?)'} = \@wozoneopt;
			}else{
				push @{$lane->{'lane_opt(s?)'}}, new RNDF::lane_opt(__STRING1__ => "zone_lane");
			}
			redraw(0,1,0);
		}elsif($cmd eq "drag-splines"){
			my $a = $window_zoom*($event->x - $start_click[0]);
			my $b = $window_zoom*($event->y - $start_click[1]);
			print "moving splines by $a $b\n";
			if(abs($a)>abs($b)){
				$splines_offx += $a;
			}else{
				$splines_offy += $b;
			}
			redraw(0,1,0);
		}elsif($cmd eq "toggle-optional-waypoint"){
			save_tree_for_undo();
			my $wp = $tree->getWaypointByWaypoint($closest_to_click);
			unless ($wp->inserted or $godmode){
				warn "Cannot mark waypoints from original RNDF file as optional";
				return;
			}
			if($wp->{comment} =~ /optional/i){
				$wp->{comment} =~ s/optional//ig;
			}else{
				$wp->{comment} .= "optional";
			}
			redraw(0,1,0);
		}elsif($cmd eq "insert-waypoint"){
			save_tree_for_undo();
			my ($ex,$ey) = win2largeimg($event->x,$event->y);
			my ($gx,$gy) = GPST12N::img2gps(\%supportpoints,\%waypoints,$ex,$ey);
			my ($hx,$hy) = @{$waypoints{$closest_to_click}};

			#### find the line which is closest to the click ...
			my $line = undef;
			my $proj = undef;
			my $dist = 1E6;
			my $wp   = V($gx,$gy);
			foreach my $l (@lines){
				die "no such waypoint: $$l[0]" unless($waypoints{$$l[0]});
				die "no such waypoint: $$l[1]" unless($waypoints{$$l[1]});
				die "waypoint $$l[0] connects to itself" if ($waypoints{$$l[0]} eq $waypoints{$$l[1]});
				my $A = V(@{$waypoints{$$l[0]}});
				my $B = V(@{$waypoints{$$l[1]}});
				my $C = V($wp->PerpFoot($A,$B));   # Lot von wp auf AB faellen

				# is C _between_ A and B?
				if (V($C->Minus($A))->Magnitude() + V($C->Minus($B))->Magnitude() < 
					V($A->Minus($B))->Magnitude() + 0.0000001){
					# is C better than the previous best?
					my $d = (V($C->Minus($wp))->Magnitude());
					if ($dist > $d){
						$line = $l;
						$proj = $C;
						$dist = $d;
					}
				}
			}
			my $lane = $tree->getLaneByWaypoint($$line[0]);
			#### Selected line: $lane->id
			my $rename = $lane->addWayPoint($gx,$gy);
			$tree->rename($rename);
			renameStuffInTree($rename);
			$lane->applyChange;
			redraw(0,1,1);
		}elsif($cmd eq "mouse-zoom"){
			save_view();
			my ($ex,$ey) = win2largeimg($event->x,$event->y);
			($window_offx,$window_offy) = win2largeimg(@start_click);
			$window_zoom = max(
				abs($ex-$window_offx)/$window_width,
				abs($ey-$window_offy)/$window_height);
			$window_zoom = int($window_zoom) unless ($window_zoom<1);
			redraw(1,1,0);
		}elsif($cmd eq "mouse-pan"){
			save_view();
			my ($ex,$ey) = win2largeimg($event->x,$event->y);
			my ($sx,$sy) = win2largeimg(@start_click);
			$window_offx = $window_offx + ($ex-$sx);
			$window_offy = $window_offy - ($ey-$sy);
			$window_offx = max(0,$window_offx);
			$window_offy = max(0,$window_offy);
			$window_offx = min($large_width-$window_offx,$window_offx);
			$window_offy = min($large_height-$window_offy,$window_offy);
			redraw(1,1,0);
		}else{
			print STDERR "Command $cmd ($action) not implemented\n";
		}
	}
	return 1;
};


package main; # <<<1

if(scalar @ARGV <2 or !-e $ARGV[0] or !-e $ARGV[1]){
	print "\n\nusage: $0 [RNDF-file] [Image-file]\n";
	exit(0);
}

my $rndffile   = shift @ARGV ;
my $rndftree   = RndfParser::get($rndffile);
my $bg         = shift @ARGV ;

RndfEditor::run($bg,$rndftree,$rndffile);
